@@ -23,6 +23,39 @@
 #include <string.h>
 #include <ctype.h>
 #include <sys/stat.h>
+#include <limits.h>
+
+
+const struct vpn_config invalid_cfg = {
+	.gateway_host = {'\0'},
+	.gateway_port = 0,
+	.username = {'\0'},
+	.password = NULL,
+	.otp = {'\0'},
+	.realm = {'\0'},
+	.set_routes = -1,
+	.set_dns = -1,
+	.pppd_use_peerdns = -1,
+	.use_syslog = -1,
+	.half_internet_routes = -1,
+	.persistent = -1,
+#if HAVE_USR_SBIN_PPPD
+	.pppd_log = NULL,
+	.pppd_plugin = NULL,
+	.pppd_ipparam = NULL,
+	.pppd_ifname = NULL,
+	.pppd_call = NULL,
+#endif
+#if HAVE_USR_SBIN_PPP
+	.ppp_system = NULL,
+#endif
+	.ca_file = NULL,
+	.user_cert = NULL,
+	.user_key = NULL,
+	.insecure_ssl = -1,
+	.cipher_list = NULL,
+	.cert_whitelist = NULL
+};
 
 /*
  * Adds a sha256 digest to the list of trusted certificates.
@@ -135,19 +168,19 @@ int load_config(struct vpn_config *cfg, const char *filename)
 		val = equals + 1;
 
 		// Remove heading spaces
-		while (key[0] != '\0' && (key[0] == ' ' || key[0] == '\t'))
+		while (isspace(key[0]))
 			key++;
-		while (val[0] != '\0' && (val[0] == ' ' || val[0] == '\t'))
+		while (isspace(val[0]))
 			val++;
 		// Remove trailing spaces
 		for (i = strlen(key) - 1; i > 0; i--) {
-			if (key[i] == ' ' || key[i] == '\t')
+			if (isspace(key[i]))
 				key[i] = '\0';
 			else
 				break;
 		}
 		for (i = strlen(val) - 1; i > 0; i--) {
-			if (val[i] == ' ' || val[i] == '\t')
+			if (isspace(val[i]))
 				val[i] = '\0';
 			else
 				break;
@@ -168,8 +201,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			strncpy(cfg->username, val, FIELD_SIZE - 1);
 			cfg->username[FIELD_SIZE] = '\0';
 		} else if (strcmp(key, "password") == 0) {
-			strncpy(cfg->password, val, FIELD_SIZE - 1);
-			cfg->password[FIELD_SIZE] = '\0';
+			cfg->password = strdup(val);
 		} else if (strcmp(key, "otp") == 0) {
 			strncpy(cfg->otp, val, FIELD_SIZE - 1);
 			cfg->otp[FIELD_SIZE] = '\0';
@@ -195,27 +227,53 @@ int load_config(struct vpn_config *cfg, const char *filename)
 		} else if (strcmp(key, "half-internet-routes") == 0) {
 			int half_internet_routes = strtob(val);
 			if (half_internet_routes < 0) {
-				log_warn("Bad half-internet-routes in config" \
-				         " file: \"%s\".\n", val);
+				log_warn("Bad half-internet-routes in config file: \"%s\".\n",
+				         val);
 				continue;
 			}
 			cfg->half_internet_routes = half_internet_routes;
+		} else if (strcmp(key, "persistent") == 0) {
+			long int persistent = strtol(val, NULL, 0);
+			if (persistent < 0 || persistent > UINT_MAX) {
+				log_warn("Bad value for persistent in config file: \"%s\".\n",
+				         val);
+				continue;
+			}
+			cfg->persistent = persistent;
+#if HAVE_USR_SBIN_PPPD
 		} else if (strcmp(key, "pppd-use-peerdns") == 0) {
 			int pppd_use_peerdns = strtob(val);
 			if (pppd_use_peerdns < 0) {
-				log_warn("Bad pppd-use-peerdns in config file:"
-				         " \"%s\".\n", val);
+				log_warn("Bad pppd-use-peerdns in config file: \"%s\".\n",
+				         val);
 				continue;
 			}
 			cfg->pppd_use_peerdns = pppd_use_peerdns;
 		} else if (strcmp(key, "pppd-log") == 0) {
+			free(cfg->pppd_log);
 			cfg->pppd_log = strdup(val);
 		} else if (strcmp(key, "pppd-plugin") == 0) {
+			free(cfg->pppd_plugin);
 			cfg->pppd_plugin = strdup(val);
 		} else if (strcmp(key, "pppd-ipparam") == 0) {
+			free(cfg->pppd_ipparam);
 			cfg->pppd_ipparam = strdup(val);
 		} else if (strcmp(key, "pppd-ifname") == 0) {
+			free(cfg->pppd_ifname);
 			cfg->pppd_ifname = strdup(val);
+		} else if (strcmp(key, "pppd-call") == 0) {
+			free(cfg->pppd_call);
+			cfg->pppd_call = strdup(val);
+#else
+		} else if (strcmp(key, "pppd") == 0) {
+			log_warn("Ignoring pppd option \"%s\".\n", key);
+#endif
+		} else if (strcmp(key, "ppp-system") == 0) {
+#if HAVE_USR_SBIN_PPP
+			cfg->ppp_system = strdup(val);
+#else
+			log_warn("Ignoring option \"%s\".\n", key);
+#endif
 		} else if (strcmp(key, "use-syslog") == 0) {
 			int use_syslog = strtob(val);
 			if (use_syslog < 0) {
@@ -226,19 +284,21 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->use_syslog = use_syslog;
 		} else if (strcmp(key, "trusted-cert") == 0) {
 			if (strlen(val) != SHA256STRLEN - 1) {
-				log_warn("Bad certificate sha256 digest in "
-				         "config file: \"%s\".\n", val);
+				log_warn("Bad certificate sha256 digest in config file: \"%s\".\n",
+				         val);
 				continue;
 			}
 			if (add_trusted_cert(cfg, val))
-				log_warn("Could not add certificate digest to "
-				         "whitelist.\n");
+				log_warn("Could not add certificate digest to whitelist.\n");
 
 		} else if (strcmp(key, "ca-file") == 0) {
+			free(cfg->ca_file);
 			cfg->ca_file = strdup(val);
 		} else if (strcmp(key, "user-cert") == 0) {
+			free(cfg->user_cert);
 			cfg->user_cert = strdup(val);
 		} else if (strcmp(key, "user-key") == 0) {
+			free(cfg->user_key);
 			cfg->user_key = strdup(val);
 		} else if (strcmp(key, "insecure-ssl") == 0) {
 			int insecure_ssl = strtob(val);
@@ -249,6 +309,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			}
 			cfg->insecure_ssl = insecure_ssl;
 		} else if (strcmp(key, "cipher-list") == 0) {
+			free(cfg->cipher_list);
 			cfg->cipher_list = strdup(val);
 		} else {
 			log_warn("Bad key in config file: \"%s\".\n", key);
@@ -265,3 +326,109 @@ err_close:
 
 	return ret;
 }
+
+void destroy_vpn_config(struct vpn_config *cfg)
+{
+	free(cfg->password);
+#if HAVE_USR_SBIN_PPPD
+	free(cfg->pppd_log);
+	free(cfg->pppd_plugin);
+	free(cfg->pppd_ipparam);
+	free(cfg->pppd_ifname);
+	free(cfg->pppd_call);
+#endif
+#if HAVE_USR_SBIN_PPP
+	free(cfg->ppp_system);
+#endif
+	free(cfg->ca_file);
+	free(cfg->user_cert);
+	free(cfg->user_key);
+	free(cfg->cipher_list);
+	while (cfg->cert_whitelist != NULL) {
+		struct x509_digest *tmp = cfg->cert_whitelist->next;
+		free(cfg->cert_whitelist);
+		cfg->cert_whitelist = tmp;
+	}
+}
+
+void merge_config(struct vpn_config *dst, struct vpn_config *src)
+{
+	if (src->gateway_host[0])
+		strcpy(dst->gateway_host, src->gateway_host);
+	if (src->gateway_port != invalid_cfg.gateway_port)
+		dst->gateway_port = src->gateway_port;
+	if (src->username[0])
+		strcpy(dst->username, src->username);
+	if (src->password != NULL && src->password[0])
+		dst->password = strdup(src->password);
+	if (src->otp[0])
+		strcpy(dst->otp, src->otp);
+	if (src->realm[0])
+		strcpy(dst->realm, src->realm);
+	if (src->set_routes != invalid_cfg.set_routes)
+		dst->set_routes = src->set_routes;
+	if (src->set_dns != invalid_cfg.set_dns)
+		dst->set_dns = src->set_dns;
+	if (src->pppd_use_peerdns != invalid_cfg.pppd_use_peerdns)
+		dst->pppd_use_peerdns = src->pppd_use_peerdns;
+	if (src->use_syslog != invalid_cfg.use_syslog)
+		dst->use_syslog = src->use_syslog;
+	if (src->half_internet_routes != invalid_cfg.half_internet_routes)
+		dst->half_internet_routes = src->half_internet_routes;
+	if (src->persistent != invalid_cfg.persistent)
+		dst->persistent = src->persistent;
+#if HAVE_USR_SBIN_PPPD
+	if (src->pppd_log) {
+		free(dst->pppd_log);
+		dst->pppd_log = src->pppd_log;
+	}
+	if (src->pppd_plugin) {
+		free(dst->pppd_plugin);
+		dst->pppd_plugin = src->pppd_plugin;
+	}
+	if (src->pppd_ipparam) {
+		free(dst->pppd_ipparam);
+		dst->pppd_ipparam = src->pppd_ipparam;
+	}
+	if (src->pppd_ifname) {
+		free(dst->pppd_ifname);
+		dst->pppd_ifname = src->pppd_ifname;
+	}
+	if (src->pppd_call) {
+		free(dst->pppd_call);
+		dst->pppd_call = src->pppd_call;
+	}
+#endif
+#if HAVE_USR_SBIN_PPP
+	if (src->ppp_system) {
+		free(dst->ppp_system);
+		dst->ppp_system = src->ppp_system;
+	}
+#endif
+	if (src->ca_file) {
+		free(dst->ca_file);
+		dst->ca_file = src->ca_file;
+	}
+	if (src->user_cert) {
+		free(dst->user_cert);
+		dst->user_cert = src->user_cert;
+	}
+	if (src->user_key) {
+		free(dst->user_key);
+		dst->user_key = src->user_key;
+	}
+	if (src->insecure_ssl != invalid_cfg.insecure_ssl)
+		dst->insecure_ssl = src->insecure_ssl;
+	if (src->cipher_list) {
+		free(dst->cipher_list);
+		dst->cipher_list = src->cipher_list;
+	}
+	if (src->cert_whitelist) {
+		while (dst->cert_whitelist != NULL) {
+			struct x509_digest *tmp = dst->cert_whitelist->next;
+			free(dst->cert_whitelist);
+			dst->cert_whitelist = tmp;
+		}
+		dst->cert_whitelist = src->cert_whitelist;
+	}
+}
