@@ -42,6 +42,12 @@ static inline const char *err_cfg_str(int code)
 	return "unknown";
 }
 
+#if HAVE_USR_SBIN_PPPD
+#define PPP_DAEMON "pppd"
+#else
+#define PPP_DAEMON "ppp"
+#endif
+
 #define SHA256LEN	(256 / 8)
 #define SHA256STRLEN	(2 * SHA256LEN + 1)
 
@@ -54,30 +60,36 @@ struct x509_digest {
 #define COOKIE_SIZE	4096
 
 struct vpn_config {
-	char 		gateway_host[FIELD_SIZE + 1];
+	char		gateway_host[FIELD_SIZE + 1];
 	struct in_addr	gateway_ip;
 	uint16_t	gateway_port;
 	char		username[FIELD_SIZE + 1];
-	char		password[FIELD_SIZE + 1];
+	char		*password;
 	char		otp[FIELD_SIZE + 1];
-	char		cookie[COOKIE_SIZE + 1];
 	char		realm[FIELD_SIZE + 1];
 
 	int	set_routes;
 	int	set_dns;
-	int     pppd_use_peerdns;
-	int     use_syslog;
+	int	pppd_use_peerdns;
+	int	use_syslog;
 	int	half_internet_routes;
 
+	unsigned int	persistent;
+
+#if HAVE_USR_SBIN_PPPD
 	char	*pppd_log;
 	char	*pppd_plugin;
 	char	*pppd_ipparam;
 	char	*pppd_ifname;
+	char	*pppd_call;
+#endif
+#if HAVE_USR_SBIN_PPP
+	char	*ppp_system;
+#endif
 
-	char	                *ca_file;
-	char	                *user_cert;
-	char	                *user_key;
-	int			verify_cert;
+	char			*ca_file;
+	char			*user_cert;
+	char			*user_key;
 	int			insecure_ssl;
 	char			*cipher_list;
 	struct x509_digest	*cert_whitelist;
@@ -87,5 +99,15 @@ int add_trusted_cert(struct vpn_config *cfg, const char *digest);
 int strtob(const char *str);
 
 int load_config(struct vpn_config *cfg, const char *filename);
+void destroy_vpn_config(struct vpn_config *cfg);
+
+/** merge source config into dest
+ *
+ * memory allocated dynamically is transferred with this function
+ * e.g. ownership goes to dest config
+ */
+void merge_config(struct vpn_config *dest, struct vpn_config *source);
+
+extern const struct vpn_config invalid_cfg;
 
 #endif
