@@ -47,11 +47,11 @@ static void url_encode(char *dest, const char *str)
 		else if (*str == ' ')
 			*dest++ = '+';
 		else {
-			static const char hex[] = "0123456789abcdef";
+			static const char hex[] = "0123456789ABCDEF";
 
 			*dest++ = '%';
-			*dest++ = hex[*str >> 4];
-			*dest++ = hex[*str & 15];
+			*dest++ = hex[(unsigned char)*str >> 4];
+			*dest++ = hex[(unsigned char)*str & 15];
 		}
 		str++;
 	}
@@ -81,6 +81,8 @@ int http_send(struct tunnel *tunnel, const char *request, ...)
 	else if (length >= BUFSZ)
 		return ERR_HTTP_TOO_LONG;
 
+	log_debug_details("%s: \n%s\n", __func__, buffer);
+
 	while (n == 0)
 		n = safe_ssl_write(tunnel->ssl_handle, (uint8_t *) buffer,
 		                   length);
@@ -93,15 +95,21 @@ int http_send(struct tunnel *tunnel, const char *request, ...)
 	return 1;
 }
 
-static const char *find_header(const char *res, const char *header)
+static const char *find_header(
+        const char *res,
+        const char *header,
+        uint32_t response_size
+)
 {
 	const char *line = res;
 
 	while (memcmp(line, "\r\n", 2)) {
-		int line_len = (char *) memmem(line, BUFSZ, "\r\n", 2) - line;
+		int line_len = (char *) memmem(
+		                       line, response_size - (line - res), "\r\n", 2
+		               ) - line;
 		int head_len = strlen(header);
 
-		if (line_len > head_len && !strncasecmp(line, header, head_len))
+		if (line_len >= head_len && !strncasecmp(line, header, head_len))
 			return line + head_len;
 		line += line_len + 2;
 	}
@@ -118,8 +126,13 @@ static const char *find_header(const char *res, const char *header)
  * @return     1         in case of success
  *             < 0       in case of error
  */
-int http_receive(struct tunnel *tunnel, char **response)
+int http_receive(
+        struct tunnel *tunnel,
+        char **response,
+        uint32_t *response_size
+)
 {
+	uint32_t res_size = BUFSZ;
 	char *buffer, *res;
 	int n = 0;
 	int bytes_read = 0;
@@ -127,7 +140,7 @@ int http_receive(struct tunnel *tunnel, char **response)
 	int content_size = 0;
 	int chunked = 0;
 
-	buffer = malloc(BUFSZ);
+	buffer = malloc(res_size);
 	if (buffer == NULL)
 		return ERR_HTTP_NO_MEM;
 
@@ -136,6 +149,7 @@ int http_receive(struct tunnel *tunnel, char **response)
 		                  (uint8_t *) buffer + bytes_read,
 		                  BUFSZ - 1 - bytes_read);
 		if (n > 0) {
+			log_debug_details("%s: \n%s\n", __func__, buffer);
 			const char *eoh;
 
 			bytes_read += n;
@@ -146,13 +160,16 @@ int http_receive(struct tunnel *tunnel, char **response)
 				if (eoh) {
 					const char *header;
 
-					header = find_header(buffer, "Content-Length: ");
+					header = find_header(
+					                 buffer,
+					                 "Content-Length: ", BUFSZ);
 					header_size = eoh - buffer + 4;
 					if (header)
 						content_size = atoi(header);
 
 					if (find_header(buffer,
-					                "Transfer-Encoding: chunked"))
+					                "Transfer-Encoding: chunked",
+					                BUFSZ))
 						chunked = 1;
 				}
 			}
@@ -200,7 +217,8 @@ int http_receive(struct tunnel *tunnel, char **response)
 		return 1;
 	}
 
-	res = realloc(buffer, bytes_read + 1);
+	res_size = bytes_read + 1;
+	res = realloc(buffer, res_size);
 	if (res == NULL) {
 		free(buffer);
 		return ERR_HTTP_NO_MEM;
@@ -208,18 +226,25 @@ int http_receive(struct tunnel *tunnel, char **response)
 	res[bytes_read] = '\0';
 
 	*response = res;
+	if (response_size != NULL)
+		*response_size = res_size;
 	return 1;
 }
 
-static int do_http_request(struct tunnel *tunnel, const char *method,
-                           const char *uri, const char *data, char **response)
+static int do_http_request(struct tunnel *tunnel,
+                           const char *method,
+                           const char *uri,
+                           const char *data,
+                           char **response,
+                           uint32_t *response_size
+                          )
 {
 	int ret;
 	const char *template = ("%s %s HTTP/1.1\r\n"
 	                        "Host: %s:%d\r\n"
 	                        "User-Agent: Mozilla/5.0 SV1\r\n"
 	                        "Accept: text/plain\r\n"
-	                        "Accept-Encoding: identify\r\n"
+	                        "Accept-Encoding: identity\r\n"
 	                        "Content-Type: application/x-www-form-urlencoded\r\n"
 	                        "Cookie: %s\r\n"
 	                        "Content-Length: %d\r\n"
@@ -227,12 +252,12 @@ static int do_http_request(struct tunnel *tunnel, const char *method,
 
 	ret = http_send(tunnel, template, method, uri,
 	                tunnel->config->gateway_host,
-	                tunnel->config->gateway_port, tunnel->config->cookie,
+	                tunnel->config->gateway_port, tunnel->cookie,
 	                strlen(data), data);
 	if (ret != 1)
 		return ret;
 
-	return http_receive(tunnel, response);
+	return http_receive(tunnel, response, response_size);
 }
 /*
  * Sends and receives data from the HTTP server.
@@ -244,13 +269,21 @@ static int do_http_request(struct tunnel *tunnel, const char *method,
  *             < 0       in case of error
  */
 static int http_request(struct tunnel *tunnel, const char *method,
-                        const char *uri, const char *data, char **response)
+                        const char *uri,
+                        const char *data,
+                        char **response,
+                        uint32_t *response_size
+                       )
 {
-	int ret = do_http_request(tunnel, method, uri, data, response);
+	int ret = do_http_request(
+	                  tunnel, method, uri, data, response, response_size);
 
 	if (ret == ERR_HTTP_SSL) {
 		ssl_connect(tunnel);
-		ret = do_http_request(tunnel, method, uri, data, response);
+		ret = do_http_request(
+		              tunnel, method, uri, data, response,
+		              response_size
+		      );
 	}
 
 	if (ret != 1)
@@ -302,14 +335,18 @@ end:
 	return ret;
 }
 
-static int get_auth_cookie(struct tunnel *tunnel, char *buf)
+static int get_auth_cookie(
+        struct tunnel *tunnel,
+        char *buf,
+        uint32_t buffer_size
+)
 {
 	int ret = 0;
 	const char *line;
 
 	ret = ERR_HTTP_NO_COOKIE;
 
-	line = find_header(buf, "Set-Cookie: ");
+	line = find_header(buf, "Set-Cookie: ", buffer_size);
 	if (line) {
 		if (strncmp(line, "SVPNCOOKIE=", 11) == 0) {
 			if (line[11] == ';' || line[11] == '\0') {
@@ -322,8 +359,8 @@ static int get_auth_cookie(struct tunnel *tunnel, char *buf)
 				if (end != NULL)
 					end[0] = '\0';
 				log_debug("Cookie: %s\n", line);
-				strncpy(tunnel->config->cookie, line, COOKIE_SIZE);
-				tunnel->config->cookie[COOKIE_SIZE] = '\0';
+				strncpy(tunnel->cookie, line, COOKIE_SIZE);
+				tunnel->cookie[COOKIE_SIZE] = '\0';
 				if (strlen(line) > COOKIE_SIZE) {
 					log_error("Cookie larger than expected: %zu > %d\n",
 					          strlen(line), COOKIE_SIZE);
@@ -337,7 +374,12 @@ static int get_auth_cookie(struct tunnel *tunnel, char *buf)
 }
 
 static
-int try_otp_auth(struct tunnel *tunnel, const char *buffer, char **res)
+int try_otp_auth(
+        struct tunnel *tunnel,
+        const char *buffer,
+        char **res,
+        uint32_t *response_size
+)
 {
 	char data[256];
 	char path[40];
@@ -458,6 +500,16 @@ int try_otp_auth(struct tunnel *tunnel, const char *buffer, char **res)
 				return -1;
 			url_encode(d, cfg->otp);
 			d += strlen(d);
+			/*  realm workaround */
+			if (cfg->realm[0] != '\0') {
+				l = strlen(cfg->realm);
+				if (!SPACE_AVAILABLE(3 * l + 8))
+					return -1;
+				strcat(d, "&realm=");
+				d += strlen(d);
+				url_encode(d, cfg->realm);
+				d += strlen(d);
+			}
 		} else if (strncmp(t, "submit", 6) == 0) {
 			/* avoid adding another '&' */
 			n = v = e = NULL;
@@ -466,7 +518,7 @@ int try_otp_auth(struct tunnel *tunnel, const char *buffer, char **res)
 	if (!SPACE_AVAILABLE(1))
 		return -1;
 	*d++ = '\0';
-	return http_request(tunnel, "POST", path, data, res);
+	return http_request(tunnel, "POST", path, data, res, response_size);
 #undef SPACE_AVAILABLE
 }
 
@@ -485,35 +537,40 @@ int auth_log_in(struct tunnel *tunnel)
 	char reqid[32] = { '\0' };
 	char polid[32] = { '\0' };
 	char group[128] = { '\0' };
-	char data[256], token[128], tokenresponse[256];
+	char data[1024], token[128], tokenresponse[256];
 	char *res = NULL;
+	uint32_t response_size;
 
 	url_encode(username, tunnel->config->username);
 	url_encode(password, tunnel->config->password);
 	url_encode(realm, tunnel->config->realm);
 
-	tunnel->config->cookie[0] = '\0';
+	tunnel->cookie[0] = '\0';
 
-	snprintf(data, 256, "username=%s&credential=%s&realm=%s&ajax=1"
+	snprintf(data, sizeof(data), "username=%s&credential=%s&realm=%s&ajax=1"
 	         "&redir=%%2Fremote%%2Findex&just_logged_in=1",
 	         username, password, realm);
 
-	ret = http_request(tunnel, "POST", "/remote/logincheck", data, &res);
+	ret = http_request(
+	              tunnel, "POST", "/remote/logincheck", data, &res, &response_size);
 	if (ret != 1)
 		goto end;
 
 	/* Probably one-time password required */
 	if (strncmp(res, "HTTP/1.1 401 Authorization Required\r\n", 37) == 0) {
-		ret = try_otp_auth(tunnel, res, &res);
+		ret = try_otp_auth(tunnel, res, &res, &response_size);
 		if (ret != 1)
 			goto end;
 	}
 
 	if (strncmp(res, "HTTP/1.1 200 OK\r\n", 17)) {
-		ret = ERR_HTTP_BAD_RES_CODE;
+		char word[17];
+
+		if (sscanf(res, "%16s %d", word, &ret) < 2)
+			ret = ERR_HTTP_BAD_RES_CODE;
 		goto end;
 	}
-	ret = get_auth_cookie(tunnel, res);
+	ret = get_auth_cookie(tunnel, res, response_size);
 	if (ret == ERR_HTTP_NO_COOKIE) {
 		struct vpn_config *cfg = tunnel->config;
 
@@ -550,20 +607,26 @@ int auth_log_in(struct tunnel *tunnel)
 		}
 
 		url_encode(tokenresponse, cfg->otp);
-		snprintf(data, 256, "username=%s&realm=%s&reqid=%s&polid=%s&grp=%s"
-		         "&code=%s&code2=&redir=%%2Fremote%%2Findex&just_logged_in=1",
+		snprintf(data, sizeof(data), "username=%s&realm=%s&reqid=%s"
+		         "&polid=%s&grp=%s&code=%s&code2="
+		         "&redir=%%2Fremote%%2Findex&just_logged_in=1",
 		         username, realm, reqid, polid, group, tokenresponse);
 
-		ret = http_request(tunnel, "POST", "/remote/logincheck", data, &res);
+		ret = http_request(
+		              tunnel, "POST", "/remote/logincheck",
+		              data, &res, &response_size);
 		if (ret != 1)
 			goto end;
 
 		if (strncmp(res, "HTTP/1.1 200 OK\r\n", 17)) {
-			ret = ERR_HTTP_BAD_RES_CODE;
+			char word[17];
+
+			if (sscanf(res, "%16s %d", word, &ret) < 2)
+				ret = ERR_HTTP_BAD_RES_CODE;
 			goto end;
 		}
 
-		ret = get_auth_cookie(tunnel, res);
+		ret = get_auth_cookie(tunnel, res, response_size);
 	}
 
 end:
@@ -573,25 +636,31 @@ end:
 
 int auth_log_out(struct tunnel *tunnel)
 {
-	return http_request(tunnel, "GET", "/remote/logout", "", NULL);
+	return http_request(tunnel, "GET", "/remote/logout", "", NULL, NULL);
 }
 
 int auth_request_vpn_allocation(struct tunnel *tunnel)
 {
-	int ret = http_request(tunnel, "GET", "/remote/index", "", NULL);
+	int ret = http_request(tunnel, "GET", "/remote/index", "", NULL, NULL);
 	if (ret != 1)
 		return ret;
 
-	return http_request(tunnel, "GET", "/remote/fortisslvpn", "", NULL);
+	return http_request(tunnel, "GET", "/remote/fortisslvpn", "", NULL, NULL);
 }
 
 static int parse_xml_config(struct tunnel *tunnel, const char *buffer)
 {
 	const char *val;
 	char *gateway;
+	int ret = 0;
 
-	if (strncmp(buffer, "HTTP/1.1 200 OK\r\n", 17))
-		return ERR_HTTP_BAD_RES_CODE;
+	if (strncmp(buffer, "HTTP/1.1 200 OK\r\n", 17)) {
+		char word[17];
+
+		if (sscanf(buffer, "%16s %d", word, &ret) < 2)
+			ret = ERR_HTTP_BAD_RES_CODE;
+		return ret;
+	}
 
 	// Skip the HTTP header
 	buffer = strstr(buffer, "\r\n\r\n");
@@ -685,7 +754,7 @@ int auth_get_config(struct tunnel *tunnel)
 	char *buffer;
 	int ret;
 
-	ret = http_request(tunnel, "GET", "/remote/fortisslvpn_xml", "", &buffer);
+	ret = http_request(tunnel, "GET", "/remote/fortisslvpn_xml", "", &buffer, NULL);
 	if (ret == 1) {
 		ret = parse_xml_config(tunnel, buffer);
 		free(buffer);
@@ -693,7 +762,7 @@ int auth_get_config(struct tunnel *tunnel)
 	if (ret == 1)
 		return ret;
 
-	ret = http_request(tunnel, "GET", "/remote/fortisslvpn", "", &buffer);
+	ret = http_request(tunnel, "GET", "/remote/fortisslvpn", "", &buffer, NULL);
 	if (ret == 1) {
 		ret = parse_config(tunnel, buffer);
 		free(buffer);
