@@ -30,7 +30,9 @@
 #include <string.h>
 #include <stdint.h>
 #include <errno.h>
+#include <limits.h>
 
+#define IPV4_GET_ROUTE_BUFFER_CHUNK_SIZE 65536
 #define SHOW_ROUTE_BUFFER_SIZE 128
 
 static char show_route_buffer[SHOW_ROUTE_BUFFER_SIZE];
@@ -72,9 +74,10 @@ static char *ipv4_show_route(struct rtentry *route)
 		strcat(show_route_buffer, " via ");
 		strncat(show_route_buffer, inet_ntoa(route_gtw(route)), 15);
 	}
-	if (route_iface(route)[0] != '\0') {
+	if (route_iface(route) != NULL) {
 		strcat(show_route_buffer, " dev ");
-		strncat(show_route_buffer, route_iface(route), ROUTE_IFACE_LEN - 1);
+		strncat(show_route_buffer, route_iface(route),
+		        SHOW_ROUTE_BUFFER_SIZE - strlen(show_route_buffer) - 1);
 	}
 
 	return show_route_buffer;
@@ -84,11 +87,6 @@ static inline int route_init(struct rtentry *route)
 {
 	memset(route, 0, sizeof(*route));
 
-	route_iface(route) = malloc(ROUTE_IFACE_LEN);
-	if (route_iface(route) == NULL)
-		return ERR_IPV4_NO_MEM;
-	route_iface(route)[0] = '\0';
-
 	cast_addr(&(route)->rt_dst)->sin_family = AF_INET;
 	cast_addr(&(route)->rt_genmask)->sin_family = AF_INET;
 	cast_addr(&(route)->rt_gateway)->sin_family = AF_INET;
@@ -98,68 +96,146 @@ static inline int route_init(struct rtentry *route)
 
 static inline void route_destroy(struct rtentry *route)
 {
-	if (route_iface(route) != NULL) {
-		free(route_iface(route));
-		route_iface(route) = NULL;
-	}
+	free(route_iface(route));
+	route_iface(route) = NULL;
 }
 
 /*
  * Finds system IP route to a destination.
  *
- * The passed route must have dest and mask set. If the route is found, the
- * function fills the gtw and iface properties.
+ * The passed route must have dest and mask set. If the route is found,
+ * the function searches for a match in the routing table and returns
+ * that one. Note that dest and mask contain the network address and
+ * the mask of the corresponding routing table entry after calling.
+ * After calling ipv4_get_route it might be necessary to set dest
+ * and mask again to the desired values for further processing.
  */
 static int ipv4_get_route(struct rtentry *route)
 {
-	size_t size;
-	char buffer[0x1000];
+	size_t buffer_size = IPV4_GET_ROUTE_BUFFER_CHUNK_SIZE;
+	char *buffer = malloc(buffer_size);
+	char *realloc_buffer;
+	int err = 0;
 	char *start, *line;
 	char *saveptr1 = NULL, *saveptr2 = NULL;
+	uint32_t rtdest, rtmask, rtgtw;
+	int rtfound = 0;
 
 	log_debug("ip route show %s\n", ipv4_show_route(route));
 
-#ifndef __APPLE__
+	// store what we are looking for
+	rtdest = route_dest(route).s_addr;
+	rtmask = route_mask(route).s_addr;
+	rtgtw = route_gtw(route).s_addr;
+
+
+	// initialize the output record
+	route_dest(route).s_addr = inet_addr("0.0.0.0");
+	route_mask(route).s_addr = inet_addr("0.0.0.0");
+	route_gtw(route).s_addr = inet_addr("0.0.0.0");
+
+#if HAVE_PROC_NET_ROUTE
+	/* this is not present on Mac OSX and FreeBSD */
 	int fd;
+	uint32_t total_bytes_read = 0;
+
 	// Cannot stat, mmap not lseek this special /proc file
 	fd = open("/proc/net/route", O_RDONLY);
-	if (fd == -1)
-		return ERR_IPV4_SEE_ERRNO;
-
-	size = read(fd, buffer, sizeof(buffer) - 1);
-	if (size == -1) {
-		close(fd);
-		return ERR_IPV4_SEE_ERRNO;
+	if (fd == -1) {
+		err = ERR_IPV4_SEE_ERRNO;
+		goto end;
 	}
+
+	int bytes_read;
+	while ((bytes_read = read(
+	                             fd, buffer + total_bytes_read,
+	                             buffer_size - total_bytes_read - 1)) > 0) {
+		total_bytes_read += bytes_read;
+
+		if ((buffer_size - total_bytes_read) < 1) {
+			buffer_size += IPV4_GET_ROUTE_BUFFER_CHUNK_SIZE;
+
+			realloc_buffer = realloc(buffer, buffer_size);
+			if (realloc_buffer) {
+				buffer = realloc_buffer;
+			} else {
+				err = ERR_IPV4_SEE_ERRNO;
+				goto end;
+			}
+		}
+	}
+
 	close(fd);
+
+	if (bytes_read < 0) {
+		err = ERR_IPV4_SEE_ERRNO;
+		goto end;
+	}
+
 #else
 	FILE *fp;
-	int len = sizeof(buffer) - 1;
+	uint32_t total_bytes_read = 0;
+
 	char *saveptr3 = NULL;
+	int have_ref = 0;
+	int have_use = 0;
+
+	static const char netstat_path[] = NETSTAT_PATH;
+	if (access(netstat_path, F_OK) != 0) {
+		log_error("%s: %s.\n", netstat_path, strerror(errno));
+		return 1;
+	}
+	log_debug("netstat_path: %s\n", netstat_path);
 
 	// Open the command for reading
-	fp = popen("/usr/sbin/netstat -f inet -rn", "r");
-	if (fp == NULL)
-		return ERR_IPV4_SEE_ERRNO;
+	fp = popen(NETSTAT_PATH" -f inet -rn", "r");
+	if (fp == NULL) {
+		err = ERR_IPV4_SEE_ERRNO;
+		goto end;
+	}
 
 	line = buffer;
 	// Read the output a line at a time
-	while (fgets(line, len, fp) != NULL) {
-		len -= strlen(line);
-		line += strlen(line);
+	while (fgets(line, buffer_size - total_bytes_read - 1, fp) != NULL) {
+		uint32_t bytes_read = strlen(line);
+		total_bytes_read += bytes_read;
+
+		if (bytes_read > 0 && line[bytes_read - 1] != '\n') {
+			buffer_size += IPV4_GET_ROUTE_BUFFER_CHUNK_SIZE;
+
+			realloc_buffer = realloc(buffer, buffer_size);
+			if (realloc_buffer) {
+				buffer = realloc_buffer;
+			} else {
+				err = ERR_IPV4_SEE_ERRNO;
+				goto end;
+			}
+		}
+
+		line = buffer + total_bytes_read;
 	}
-	size = sizeof(buffer)-1 - len;
 	pclose(fp);
 
 	// reserve enough memory (256 shorts)
 	// to make sure not to access out of bounds later,
 	// for ipv4 only unsigned short is allowed
 
-	unsigned short flag_table[256];
-	memset(flag_table, 0, 256*sizeof(short));
+	unsigned short flag_table[256] = { 0 };
+
+	/*
+	 * Fill the flag_table now. Unfortunately it is not easy
+	 * to do this in a more elegant way. The problem here
+	 * is that these are already preprocessor macros and
+	 * we can't use them as arguments for another macro which
+	 * would include the #ifdef statements.
+	 *
+	 * Also, not all flags might be allowed in the context
+	 * of ipv4, and the code depends on which ones are
+	 * actually implemented on the target platform, which
+	 * might also be varying between OSX versions.
+	 *
+	 */
 
-	// fill the table now (I'm still looking for a more elagant way to do this),
-	// also, not all flags might be allowed in the context of ipv4
 #ifdef RTF_PROTO1     // Protocol specific routing flag #1
 	flag_table['1'] = RTF_PROTO1 & USHRT_MAX;
 #endif
@@ -226,32 +302,37 @@ static int ipv4_get_route(struct rtentry *route)
 #ifdef RTF_PROXY      // Proxying; cloned routes will not be scoped
 	flag_table['Y'] = RTF_PROXY & USHRT_MAX;
 #endif
+
 #endif
 
-	if (size == 0) {
+	if (total_bytes_read == 0) {
 		log_debug("routing table is empty.\n");
-		return ERR_IPV4_PROC_NET_ROUTE;
+		err = ERR_IPV4_PROC_NET_ROUTE;
+		goto end;
 	}
-	buffer[size] = '\0';
+	buffer[total_bytes_read] = '\0';
 
 	// Skip first line
 	start = index(buffer, '\n');
 	if (start == NULL) {
 		log_debug("routing table is malformed.\n");
-		return ERR_IPV4_PROC_NET_ROUTE;
+		err = ERR_IPV4_PROC_NET_ROUTE;
+		goto end;
 	}
 	start++;
 
-#ifdef __APPLE__
-	// Skip 3 more line
+#if !HAVE_PROC_NET_ROUTE
+	if (strstr(buffer, "Ref") != NULL) have_ref = 1;
+	if (strstr(buffer, "Use") != NULL) have_use = 1;
+	// Skip 3 more lines from netstat output on Mac OSX and on FreeBSD
 	start = index(start, '\n');
 	start = index(++start, '\n');
 	start = index(++start, '\n');
 	if (start == NULL) {
 		log_debug("routing table is malformed.\n");
-		return ERR_IPV4_PROC_NET_ROUTE;
+		err = ERR_IPV4_PROC_NET_ROUTE;
+		goto end;
 	}
-
 #endif
 
 	// Look for the route
@@ -260,7 +341,7 @@ static int ipv4_get_route(struct rtentry *route)
 		char *iface;
 		uint32_t dest, mask, gtw;
 		unsigned short flags;
-#ifndef __APPLE__
+#if HAVE_PROC_NET_ROUTE
 		unsigned short irtt;
 		short metric;
 		unsigned long mtu, window;
@@ -277,19 +358,24 @@ static int ipv4_get_route(struct rtentry *route)
 		window = strtol(strtok_r(NULL, "\t", &saveptr2), NULL, 16);
 		irtt = strtol(strtok_r(NULL, "\t", &saveptr2), NULL, 16);
 #else
+		/* parse netstat output on Mac OSX and BSD */
 		char tmp_ip_string[16];
 		struct in_addr dstaddr;
 		int pos;
 		char *tmpstr;
 
-		log_debug("line: %s\n", line);
+		log_debug_details("line: %s\n", line);
 
 		saveptr3 = NULL;
 		dest = UINT32_MAX;
 		mask = UINT32_MAX;
 		// "Destination"
 		tmpstr = strtok_r(line, " ", &saveptr2);
-		log_debug("- Destination: %s\n", tmpstr);
+		if (strncmp(tmpstr, "Internet6", 9) == 0) {
+			// we have arrived at the end of ipv4 output
+			goto end;
+		}
+		log_debug_details("- Destination: %s\n", tmpstr);
 		// replace literal "default" route by IPV4 numbers-and-dots notation
 		if (strncmp(tmpstr, "default", 7) == 0) {
 			dest = 0;
@@ -339,13 +425,13 @@ static int ipv4_get_route(struct rtentry *route)
 			}
 
 		}
-		log_debug("- Destination IP Hex: %x\n", dest);
-		log_debug("- Destination Mask Hex: %x\n", mask);
+		log_debug_details("- Destination IP Hex: %x\n", dest);
+		log_debug_details("- Destination Mask Hex: %x\n", mask);
 		// "Gateway"
 		gtw = 0;
 		if (inet_aton(strtok_r(NULL, " ", &saveptr2), &dstaddr)) {
 			gtw = dstaddr.s_addr;
-			log_debug("- Gateway Mask Hex: %x\n", gtw);
+			log_debug_details("- Gateway Mask Hex: %x\n", gtw);
 		}
 		// "Flags"
 		tmpstr = strtok_r(NULL, " ", &saveptr2);
@@ -353,40 +439,115 @@ static int ipv4_get_route(struct rtentry *route)
 		// this is the reason for the 256 entries mentioned above
 		for (pos = 0; pos < strlen(tmpstr); pos++)
 			flags |= flag_table[(unsigned char)tmpstr[pos]];
-		strtok_r(NULL, " ", &saveptr2); // "Refs"
-		strtok_r(NULL, " ", &saveptr2); // "Use"
-		iface = strtok_r(NULL, " ", &saveptr2); // "Netif"
-		log_debug("- Interface: %s\n", iface);
-		log_debug("\n");
-#endif
 
-		if (dest == route_dest(route).s_addr &&
-		    mask == route_mask(route).s_addr) {
-			// Requested route has been found
-			route_gtw(route).s_addr = gtw;
-			route->rt_flags = flags;
-#ifndef __APPLE__
-			// we do not have these values from Mac OS X netstat,
-			// so stay with defaults denoted by values of 0
-			route->rt_metric = metric;
-			route->rt_mtu = mtu;
-			route->rt_window = window;
-			route->rt_irtt = irtt;
+		if (have_ref) strtok_r(NULL, " ", &saveptr2); // "Ref"
+		if (have_use) strtok_r(NULL, " ", &saveptr2); // "Use"
+
+		iface = strtok_r(NULL, " ", &saveptr2); // "Netif"
+		log_debug_details("- Interface: %s\n", iface);
+		log_debug_details("\n");
+#endif
+		/*
+		 * Now that we have parsed a routing entry, check if it
+		 * matches the current argument to the function call.
+		 * In rtentry we have integer representation, i.e.
+		 * the most significant byte corresponds to the last
+		 * number of dotted-number representation and vice versa.
+		 * In this representation ( address & mask ) is the network
+		 * address.
+		 * The routing algorithm does the following:
+		 * First, check if the network address we are looking for
+		 * falls into the network for the current route.
+		 * Therefore, calculate the network address for both, the
+		 * current route and for the destination we are searching.
+		 * If the destination is a smaller network (for instance a
+		 * single host), we have to mask again with the netmask of
+		 * the routing entry that we are checking in order to obtain
+		 * the network address in the context of the current route.
+		 * If both network addresses match, we have found a candidate
+		 * for a route.
+		 * However, there might be another route for a smaller network,
+		 * therefore repeat this and only store the resulting route
+		 * when the mask is at least as large as the one we may
+		 * have already found in a previous iteration (a larger
+		 * netmask corresponds to a smaller network in this
+		 * representation, and has a higher prority by default).
+		 * Also, only consider routing entries for which the
+		 * netmask is not larger than the netmask used in the
+		 * argument when calling the function - so that we can
+		 * distinguish between different routing entries for subnets
+		 * of different size but with the same network address.
+		 * For routing entries with the same destination and
+		 * the same netmask the metric can be used for adjusting
+		 * the priority (this is not supported on mac).
+		 * If the metric is larger than one found for this network
+		 * size, skip the current route (smaller numbers denote
+		 * less hops and therefore have a higher priority).
+		 */
+
+		if (((dest & mask) == (rtdest & rtmask & mask))
+		    && (mask >= route_mask(route).s_addr)
+		    && (mask <= rtmask)
+		    && ((route_iface(route) == NULL)
+		        || (strcmp(iface, route_iface(route)) == 0)
+		        || (strlen(route_iface(route)) > 0
+		            && route_iface(route)[0] == '!'
+		            && strcmp(iface, &route_iface(route)[1]) != 0)
+		       )) {
+#if HAVE_PROC_NET_ROUTE
+			if (((mask == route_mask(route).s_addr)
+			     && (metric <= route->rt_metric))
+			    || (rtfound == 0)
+			    || (mask > route_mask(route).s_addr)) {
+#endif
+				rtfound = 1;
+				// Requested route has been found
+				route_dest(route).s_addr = dest;
+				route_mask(route).s_addr = mask;
+				route_gtw(route).s_addr = gtw;
+				route->rt_flags = flags;
+
+				free(route_iface(route));
+				route_iface(route) = strdup(iface);
+				if (!route_iface(route))
+					return ERR_IPV4_NO_MEM;
+
+#if HAVE_PROC_NET_ROUTE
+				// we do not have these values from Mac OS X netstat,
+				// so stay with defaults denoted by values of 0
+				route->rt_metric = metric;
+				route->rt_mtu = mtu;
+				route->rt_window = window;
+				route->rt_irtt = irtt;
+			}
 #endif
-			strncpy(route_iface(route), iface,
-			        ROUTE_IFACE_LEN - 1);
-			return 0;
 		}
 		line = strtok_r(NULL, "\n", &saveptr1);
 	}
-	log_debug("Route not found.\n");
+end:
+	free(buffer);
 
-	return ERR_IPV4_NO_SUCH_ROUTE;
+	if (err)
+		return err;
+
+	if (rtfound==0) {
+		// should not occur anymore unless there is no default route
+		log_debug("Route not found.\n");
+		// at least restore input values
+		route_dest(route).s_addr = rtdest;
+		route_mask(route).s_addr = rtmask;
+		route_gtw(route).s_addr = rtgtw;
+
+		return ERR_IPV4_NO_SUCH_ROUTE;
+	}
+
+	return 0;
 }
 
 static int ipv4_set_route(struct rtentry *route)
 {
-#ifndef __APPLE__
+#ifdef HAVE_RT_ENTRY_WITH_RT_DST
+	/* we can copy rtentry struct directly between openfortivpn and kernel */
 	log_debug("ip route add %s\n", ipv4_show_route(route));
 
 	int sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
@@ -399,18 +560,33 @@ static int ipv4_set_route(struct rtentry *route)
 	}
 	close(sockfd);
 #else
+	/* we have to use the route command as tool for route manipulation */
 	char cmd[SHOW_ROUTE_BUFFER_SIZE];
 
-	strcpy(cmd, "route -n add -net ");
+	if (access("/sbin/route", F_OK) != 0) {
+		log_error("/sbin/route: %s.\n", strerror(errno));
+		return 1;
+	}
+
+
+	strcpy(cmd, "/sbin/route -n add ");
+	if (route->rt_flags & RTF_HOST)
+		strcat(cmd, "-host ");
+	else
+		strcat(cmd, "-net ");
+
 	strncat(cmd, inet_ntoa(route_dest(route)), 15);
-	strcat(cmd, " -netmask ");
-	strncat(cmd, inet_ntoa(route_mask(route)), 15);
+	if (!(route->rt_flags & RTF_HOST)) {
+		strcat(cmd, " -netmask ");
+		strncat(cmd, inet_ntoa(route_mask(route)), 15);
+	}
 	if (route->rt_flags & RTF_GATEWAY) {
 		strcat(cmd, " ");
 		strncat(cmd, inet_ntoa(route_gtw(route)), 15);
 	} else {
 		strcat(cmd, " -interface ");
-		strcat(cmd, route_iface(route));
+		strncat(cmd, route_iface(route),
+		        SHOW_ROUTE_BUFFER_SIZE - strlen(cmd) - 1);
 	}
 
 	log_debug("%s\n", cmd);
@@ -425,7 +601,8 @@ static int ipv4_set_route(struct rtentry *route)
 
 static int ipv4_del_route(struct rtentry *route)
 {
-#ifndef __APPLE__
+#ifdef HAVE_RT_ENTRY_WITH_RT_DST
+	/* we can copy rtentry struct directly between openfortivpn and kernel */
 	struct rtentry tmp;
 	int sockfd;
 
@@ -449,10 +626,31 @@ static int ipv4_del_route(struct rtentry *route)
 #else
 	char cmd[SHOW_ROUTE_BUFFER_SIZE];
 
-	strcpy(cmd, "route -n delete ");
+	if (access("/sbin/route", F_OK) != 0) {
+		log_error("/sbin/route: %s.\n", strerror(errno));
+		return 1;
+	}
+
+
+	strcpy(cmd, "/sbin/route -n delete ");
+	if (route->rt_flags & RTF_HOST)
+		strcat(cmd, "-host ");
+	else
+		strcat(cmd, "-net ");
+
 	strncat(cmd, inet_ntoa(route_dest(route)), 15);
-	strcat(cmd, " -netmask ");
-	strncat(cmd, inet_ntoa(route_mask(route)), 15);
+	if (!(route->rt_flags & RTF_HOST)) {
+		strcat(cmd, " -netmask ");
+		strncat(cmd, inet_ntoa(route_mask(route)), 15);
+	}
+	if (route->rt_flags & RTF_GATEWAY) {
+		strcat(cmd, " ");
+		strncat(cmd, inet_ntoa(route_gtw(route)), 15);
+	} else {
+		strcat(cmd, " -interface ");
+		strncat(cmd, route_iface(route),
+		        SHOW_ROUTE_BUFFER_SIZE - strlen(cmd) - 1);
+	}
 
 	log_debug("%s\n", cmd);
 
@@ -475,22 +673,38 @@ int ipv4_protect_tunnel_route(struct tunnel *tunnel)
 	// Back up default route
 	route_dest(def_rt).s_addr = inet_addr("0.0.0.0");
 	route_mask(def_rt).s_addr = inet_addr("0.0.0.0");
+	route_iface(def_rt) = malloc(strlen(tunnel->ppp_iface) + 2);
+	sprintf(route_iface(def_rt), "!%s", tunnel->ppp_iface);
 
 	ret = ipv4_get_route(def_rt);
 	if (ret != 0) {
 		log_warn("Could not get current default route (%s).\n",
 		         err_ipv4_str(ret));
-		log_warn("Protecting tunnel route has failed. "
-		         "But this can be working except for some cases.\n");
+		log_warn("Protecting tunnel route has failed. But this can be working except for some cases.\n");
 		goto err_destroy;
 	}
 
-
-	// Set the default route as the route to the tunnel gateway
-	char *iface = route_iface(gtw_rt);
-	memcpy(gtw_rt, def_rt, sizeof(*gtw_rt));
-	route_iface(gtw_rt) = iface;
-	strncpy(route_iface(gtw_rt), route_iface(def_rt), ROUTE_IFACE_LEN - 1);
+	// Set the up a route to the tunnel gateway
+	route_dest(gtw_rt).s_addr = tunnel->config->gateway_ip.s_addr;
+	route_mask(gtw_rt).s_addr = inet_addr("255.255.255.255");
+	route_iface(gtw_rt) = malloc(strlen(tunnel->ppp_iface) + 2);
+	sprintf(route_iface(gtw_rt), "%s", tunnel->ppp_iface);
+	ret = ipv4_get_route(gtw_rt);
+	if ((ret == 0)
+	    && (route_dest(gtw_rt).s_addr == tunnel->config->gateway_ip.s_addr)
+	    && (route_mask(gtw_rt).s_addr == inet_addr("255.255.255.255"))) {
+		log_debug("removing wrong route to vpn server...\n");
+		log_debug("ip route show %s\n", ipv4_show_route(gtw_rt));
+		ipv4_del_route(gtw_rt);
+	}
+	sprintf(route_iface(gtw_rt),"!%s",tunnel->ppp_iface);
+	ret = ipv4_get_route(gtw_rt);
+	if (ret != 0) {
+		log_warn("Could not get route to gateway (%s).\n",
+		         err_ipv4_str(ret));
+		log_warn("Protecting tunnel route has failed. But this can be working except for some cases.\n");
+		goto err_destroy;
+	}
 	route_dest(gtw_rt).s_addr = tunnel->config->gateway_ip.s_addr;
 	route_mask(gtw_rt).s_addr = inet_addr("255.255.255.255");
 	gtw_rt->rt_flags |= RTF_HOST;
@@ -498,6 +712,7 @@ int ipv4_protect_tunnel_route(struct tunnel *tunnel)
 
 	tunnel->ipv4.route_to_vpn_is_added = 1;
 	log_debug("Setting route to vpn server...\n");
+	log_debug("ip route show %s\n", ipv4_show_route(gtw_rt));
 	ret = ipv4_set_route(gtw_rt);
 	if (ret == ERR_IPV4_SEE_ERRNO && errno == EEXIST) {
 		log_warn("Route to vpn server exists already.\n");
@@ -515,12 +730,42 @@ err_destroy:
 	return ret;
 }
 
+
+#if HAVE_USR_SBIN_PPPD
+static void add_text_route(struct tunnel *tunnel, const char *dest,
+                           const char *mask, const char *gw)
+{
+	size_t l0, l1;
+	const char fmt[] = ",%s/%s/%s";
+	const char trigger[] = "openfortivpn";
+	char **target = &tunnel->config->pppd_ipparam;
+	char *ptr;
+
+	if (*target == NULL || strncmp(*target, trigger, strlen(trigger)))
+		return;
+	if (!dest || !mask || !gw)
+		return;
+	log_info("Registering route %s/%s via %s\n", dest, mask, gw);
+	l0 = strlen(*target);
+	l1 = strlen(fmt) + strlen(dest) + strlen(mask) + strlen(gw) + 1;
+	if ((ptr = realloc(*target, l0 + l1))) {
+		*target = ptr;
+		snprintf(*target + l0, l1, fmt, dest, mask, gw);
+	} else {
+		log_error("realloc: %s\n", strerror(errno));
+	}
+}
+#endif
+
 int ipv4_add_split_vpn_route(struct tunnel *tunnel, char *dest, char *mask,
                              char *gateway)
 {
 	struct rtentry *route;
 	char env_var[24];
 
+#if HAVE_USR_SBIN_PPPD
+	add_text_route(tunnel, dest, mask, gateway);
+#endif
 	if (tunnel->ipv4.split_routes == MAX_SPLIT_ROUTES)
 		return ERR_IPV4_NO_MEM;
 	if ((tunnel->ipv4.split_rt == NULL)
@@ -555,7 +800,10 @@ int ipv4_add_split_vpn_route(struct tunnel *tunnel, char *dest, char *mask,
 		route_gtw(route).s_addr = inet_addr(gateway);
 		route->rt_flags |= RTF_GATEWAY;
 	} else {
-		strncpy(route_iface(route), tunnel->ppp_iface, ROUTE_IFACE_LEN - 1);
+		free(route_iface(route));
+		route_iface(route) = strdup(tunnel->ppp_iface);
+		if (!route_iface(route))
+			return ERR_IPV4_NO_MEM;
 	}
 
 	return 0;
@@ -569,11 +817,18 @@ static int ipv4_set_split_routes(struct tunnel *tunnel)
 		struct rtentry *route;
 		int ret;
 		route = &tunnel->ipv4.split_rt[i];
-		strncpy(route_iface(route), tunnel->ppp_iface,
-		        ROUTE_IFACE_LEN - 1);
+		free(route_iface(route));
+		route_iface(route) = strdup(tunnel->ppp_iface);
+		if (!route_iface(route))
+			return ERR_IPV4_NO_MEM;
+#ifndef HAVE_RT_ENTRY_WITH_RT_DST
+		if (route_gtw(route).s_addr == tunnel->ipv4.ip_addr.s_addr)
+			route_gtw(route).s_addr = 0;
 		if (route_gtw(route).s_addr == 0)
-			route_gtw(route).s_addr = tunnel->ipv4.ip_addr.s_addr;
-		route->rt_flags |= RTF_GATEWAY;
+			route->rt_flags &= ~RTF_GATEWAY;
+#endif
+		if (route_gtw(route).s_addr != 0)
+			route->rt_flags |= RTF_GATEWAY;
 		ret = ipv4_set_route(route);
 		if (ret == ERR_IPV4_SEE_ERRNO && errno == EEXIST)
 			log_warn("Route to gateway exists already.\n");
@@ -608,7 +863,10 @@ static int ipv4_set_default_routes(struct tunnel *tunnel)
 		route_gtw(ppp_rt).s_addr = inet_addr("0.0.0.0");
 		log_debug("Setting new default route...\n");
 
-		strncpy(route_iface(ppp_rt), tunnel->ppp_iface, ROUTE_IFACE_LEN - 1);
+		free(route_iface(ppp_rt));
+		route_iface(ppp_rt) = strdup(tunnel->ppp_iface);
+		if (!route_iface(ppp_rt))
+			return ERR_IPV4_NO_MEM;
 
 		ret = ipv4_set_route(ppp_rt);
 		if (ret == ERR_IPV4_SEE_ERRNO && errno == EEXIST) {
@@ -626,7 +884,10 @@ static int ipv4_set_default_routes(struct tunnel *tunnel)
 		route_dest(ppp_rt).s_addr = inet_addr("0.0.0.0");
 		route_mask(ppp_rt).s_addr = inet_addr("128.0.0.0");
 
-		strncpy(route_iface(ppp_rt), tunnel->ppp_iface, ROUTE_IFACE_LEN - 1);
+		free(route_iface(ppp_rt));
+		route_iface(ppp_rt) = strdup(tunnel->ppp_iface);
+		if (!route_iface(ppp_rt))
+			return ERR_IPV4_NO_MEM;
 
 		ret = ipv4_set_route(ppp_rt);
 		if (ret == ERR_IPV4_SEE_ERRNO && errno == EEXIST) {
@@ -687,8 +948,7 @@ int ipv4_restore_routes(struct tunnel *tunnel)
 			// automatically restored on all linux distributions
 			ret = ipv4_set_route(def_rt);
 			if (ret != 0) {
-				log_warn("Could not restore default route (%s). "
-				         "Already restored?\n",
+				log_warn("Could not restore default route (%s). Already restored?\n",
 				         err_ipv4_str(ret));
 			}
 
@@ -709,7 +969,7 @@ int ipv4_add_nameservers_to_resolv_conf(struct tunnel *tunnel)
 	FILE *file;
 	struct stat stat;
 	char ns1[28], ns2[28]; // 11 + 15 + 1 + 1
-	char *buffer, *line;
+	char *buffer;
 
 	tunnel->ipv4.ns_are_new = 1;
 
@@ -738,7 +998,7 @@ int ipv4_add_nameservers_to_resolv_conf(struct tunnel *tunnel)
 	buffer = malloc(stat.st_size + 1);
 	if (buffer == NULL) {
 		log_warn("Could not read /etc/resolv.conf (%s).\n",
-		         "Not enough memory");
+		         strerror(errno));
 		goto err_close;
 	}
 
@@ -759,12 +1019,12 @@ int ipv4_add_nameservers_to_resolv_conf(struct tunnel *tunnel)
 		ns2[0] = '\0';
 	}
 
-	for (line = strtok(buffer, "\n"); line != NULL;
+	for (const char *line = strtok(buffer, "\n");
+	     line != NULL;
 	     line = strtok(NULL, "\n")) {
 		if (strcmp(line, ns1) == 0) {
 			tunnel->ipv4.ns_are_new = 0;
-			log_debug("Nameservers already present in "
-			          "/etc/resolv.conf.\n");
+			log_debug("Nameservers already present in /etc/resolv.conf.\n");
 			ret = 0;
 			goto err_free;
 		}
@@ -805,7 +1065,7 @@ int ipv4_del_nameservers_from_resolv_conf(struct tunnel *tunnel)
 	FILE *file;
 	struct stat stat;
 	char ns1[27], ns2[27]; // 11 + 15 + 1
-	char *buffer, *line;
+	char *buffer;
 
 	// If nameservers were already there before setting up tunnel,
 	// don't delete them from /etc/resolv.conf
@@ -831,7 +1091,7 @@ int ipv4_del_nameservers_from_resolv_conf(struct tunnel *tunnel)
 	buffer = malloc(stat.st_size + 1);
 	if (buffer == NULL) {
 		log_warn("Could not read /etc/resolv.conf (%s).\n",
-		         "Not enough memory");
+		         strerror(errno));
 		goto err_close;
 	}
 
@@ -855,11 +1115,13 @@ int ipv4_del_nameservers_from_resolv_conf(struct tunnel *tunnel)
 		goto err_free;
 	}
 
-	for (line = strtok(buffer, "\n"); line != NULL; line = strtok(NULL, "\n")) {
+	for (const char *line = strtok(buffer, "\n");
+	     line != NULL;
+	     line = strtok(NULL, "\n")) {
 		if (strcmp(line, ns1) == 0) {
 			log_debug("Deleting \"%s\" from /etc/resolv.conf.\n", ns1);
 		} else if (strcmp(line, ns2) == 0) {
-			log_debug("Deleting \"%s\" from /etc/resolv.conf.\n", ns1);
+			log_debug("Deleting \"%s\" from /etc/resolv.conf.\n", ns2);
 		} else {
 			fputs(line, file);
 			fputs("\n", file);
