@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <errno.h>
 
 static pthread_mutex_t mutex;
 static int do_syslog = 0;
@@ -38,7 +39,7 @@ struct log_param_s {
 	int syslog_prio;
 };
 
-static const struct log_param_s log_params[OFV_LOG_DEBUG_PACKETS + 1] = {
+static const struct log_param_s log_params[OFV_LOG_DEBUG_DETAILS + 1] = {
 	{ "        ", "",           LOG_ERR},
 	{ "ERROR:  ", "\033[0;31m", LOG_ERR},
 	{ "WARN:   ", "\033[0;33m", LOG_WARNING},
@@ -54,7 +55,7 @@ void init_logging(void)
 	is_a_tty = isatty(STDOUT_FILENO);
 
 	pthread_mutexattr_init(&mutexattr);
-#ifndef __APPLE__
+#ifdef HAVE_PTHREAD_MUTEXATTR_SETROBUST
 	pthread_mutexattr_setrobust(&mutexattr, PTHREAD_MUTEX_ROBUST);
 #endif
 	pthread_mutex_init(&mutex, &mutexattr);
@@ -70,7 +71,7 @@ void set_syslog(int use_syslog)
 
 void increase_verbosity(void)
 {
-	if (loglevel < OFV_LOG_DEBUG_PACKETS)
+	if (loglevel < OFV_LOG_DEBUG_DETAILS)
 		loglevel++;
 }
 void decrease_verbosity(void)
@@ -87,7 +88,7 @@ void do_log(int verbosity, const char *format, ...)
 	pthread_mutex_lock(&mutex);
 
 	// Use sane default if wrong verbosity specified
-	if (verbosity > OFV_LOG_DEBUG || verbosity < 0)
+	if (verbosity > OFV_LOG_DEBUG_DETAILS || verbosity < 0)
 		verbosity = OFV_LOG_MUTE;
 	lp = &log_params[verbosity];
 
@@ -118,7 +119,7 @@ void do_log_packet(const char *prefix, size_t len, const uint8_t *packet)
 
 	str = malloc(strlen(prefix) + 3 * len + 1 + 1);
 	if (str == NULL) {
-		log_error("malloc failed\n");
+		log_error("malloc: %s\n", strerror(errno));
 		return;
 	}
 
