@@ -25,27 +25,61 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <limits.h>
 
-#define usage \
-"Usage: openfortivpn [<host>:<port>] [-u <user>] [-p <pass>]\n" \
-"                    [--realm=<realm>] [--otp=<otp>] [--set-routes=<0|1>]\n" \
-"                    [--half_internet_routes=<0|1>] [--set-dns=<0|1>]\n" \
+#define PWD_BUFSIZ	4096
+
+
+#if HAVE_USR_SBIN_PPPD
+#define PPPD_USAGE \
 "                    [--pppd-no-peerdns] [--pppd-log=<file>]\n" \
 "                    [--pppd-ifname=<string>] [--pppd-ipparam=<string>]\n" \
-"                    [--pppd-plugin=<file>] [--ca-file=<file>]\n" \
+"                    [--pppd-call=<name>] [--pppd-plugin=<file>]\n"
+
+#define PPPD_HELP \
+"  --pppd-no-peerdns             Do not ask peer ppp server for DNS server addresses\n" \
+"                                and do not make pppd rewrite /etc/resolv.conf\n" \
+"  --pppd-log=<file>             Set pppd in debug mode and save its logs into\n" \
+"                                <file>.\n" \
+"  --pppd-plugin=<file>          Use specified pppd plugin instead of configuring\n" \
+"                                resolver and routes directly.\n" \
+"  --pppd-ifname=<string>        Set the pppd interface name, if supported by pppd.\n" \
+"  --pppd-ipparam=<string>       Provides  an extra parameter to the ip-up, ip-pre-up\n" \
+"                                and ip-down scripts. See man (8) pppd\n" \
+"  --pppd-call=<name>            Move most pppd options from pppd cmdline to\n" \
+"                                /etc/ppp/peers/<name> and invoke pppd with\n" \
+"                                'call <name>'\n"
+#endif
+#if HAVE_USR_SBIN_PPP
+#define PPPD_USAGE \
+"                    [--ppp-system=<system>]\n"
+#define PPPD_HELP \
+"  --ppp-system=<system>         connect to the specified system as defined in\n" \
+"                                /etc/ppp/ppp.conf\n"
+#endif
+
+#define usage \
+"Usage: openfortivpn [<host>[:<port>]] [-u <user>] [-p <pass>]\n" \
+"                    [--realm=<realm>] [--otp=<otp>] [--set-routes=<0|1>]\n" \
+"                    [--half-internet-routes=<0|1>] [--set-dns=<0|1>]\n" \
+PPPD_USAGE \
+"                    [--ca-file=<file>]\n" \
 "                    [--user-cert=<file>] [--user-key=<file>]\n" \
 "                    [--trusted-cert=<digest>] [--use-syslog]\n" \
-"                    [-c <file>] [-v|-q]\n" \
+"                    [--persistent=<interval>] [-c <file>] [-v|-q]\n" \
 "       openfortivpn --help\n" \
 "       openfortivpn --version\n" \
 "\n"
 
-#define help_options \
+#define summary \
 "Client for PPP+SSL VPN tunnel services.\n" \
 "openfortivpn connects to a VPN by setting up a tunnel to the gateway at\n" \
 "<host>:<port>. It spawns a pppd process and operates the communication between\n" \
 "the gateway and this process.\n" \
-"\n" \
+"\n"
+
+
+#define help_options \
 "Options:\n" \
 "  -h --help                     Show this help message and exit.\n" \
 "  --version                     Show version and exit.\n" \
@@ -56,11 +90,12 @@
 "  -o <otp>, --otp=<otp>         One-Time-Password.\n" \
 "  --realm=<realm>               Use specified authentication realm on VPN gateway\n" \
 "                                when tunnel is up.\n" \
-"  --set-routes=[01]             Set if we should configure output roues through\n" \
+"  --set-routes=[01]             Set if openfortivpn should configure output routes through\n" \
 "                                the VPN when tunnel is up.\n" \
 "  --no-routes                   Do not configure routes, same as --set-routes=0.\n" \
-"  --half-internet-routes=[01]   Add /1-routes instead of replacing the default route\n" \
-"  --set-dns=[01]                Set if we should add VPN name servers in\n" \
+"  --half-internet-routes=[01]   Add two 0.0.0.0/1 and 128.0.0.0/1 routes with higher\n" \
+"                                priority instead of replacing the default route.\n" \
+"  --set-dns=[01]                Set if openfortivpn should add VPN name servers in\n" \
 "                                /etc/resolv.conf\n" \
 "  --no-dns                      Do not reconfigure DNS, same as --set-dns=0\n" \
 "  --ca-file=<file>              Use specified PEM-encoded certificate bundle\n" \
@@ -68,8 +103,8 @@
 "                                certificate.\n" \
 "  --user-cert=<file>            Use specified PEM-encoded certificate if the server\n" \
 "                                requires authentication with a certificate.\n" \
-"  --user-key=<file>             Use specified PEM-encoded key if the server if the\n" \
-"                                server requires authentication with a certificate.\n" \
+"  --user-key=<file>             Use specified PEM-encoded key if the server requires\n" \
+"                                authentication with a certificate.\n" \
 "  --use-syslog                  Log to syslog instead of terminal.\n" \
 "  --trusted-cert=<digest>       Trust a given gateway. If classical SSL\n" \
 "                                certificate validation fails, the gateway\n" \
@@ -84,15 +119,9 @@
 "                                you can try with the cipher suggested in the output\n" \
 "                                of 'openssl s_client -connect <host:port>'\n" \
 "                                (e.g. AES256-GCM-SHA384)\n" \
-"  --pppd-no-peerdns             Do not ask peer ppp server for DNS addresses\n" \
-"                                and do not make pppd rewrite /etc/resolv.conf\n" \
-"  --pppd-log=<file>             Set pppd in debug mode and save its logs into\n" \
-"                                <file>.\n" \
-"  --pppd-plugin=<file>          Use specified pppd plugin instead of configuring\n" \
-"                                resolver and routes directly.\n" \
-"  --pppd-ifname=<string>        Set the pppd interface name, if supported by pppd.\n" \
-"  --pppd-ipparam=<string>       Provides  an extra parameter to the ip-up, ip-pre-up\n" \
-"                                and ip-down scripts. See man (8) pppd\n" \
+PPPD_HELP \
+"  --persistent=<interval>       Run the vpn persistently in a loop and try to re-\n" \
+"                                connect every <interval> seconds when dropping out\n" \
 "  -v                            Increase verbosity. Can be used multiple times\n" \
 "                                to be even more verbose.\n" \
 "  -q                            Decrease verbosity. Can be used multiple times\n" \
@@ -116,55 +145,44 @@
 "      trusted-cert = othercertificatedigest6631bf...\n" \
 "  For a full-featured config see man openfortivpn(1).\n"
 
-static inline void destroy_vpn_config(struct vpn_config *cfg)
-{
-	while (cfg->cert_whitelist != NULL) {
-		struct x509_digest *tmp = cfg->cert_whitelist->next;
-		free(cfg->cert_whitelist);
-		cfg->cert_whitelist = tmp;
-	}
-	free(cfg->cipher_list);
-	free(cfg->user_key);
-	free(cfg->user_cert);
-	free(cfg->ca_file);
-	free(cfg->pppd_ipparam);
-	free(cfg->pppd_plugin);
-	free(cfg->pppd_log);
-}
-
 int main(int argc, char **argv)
 {
 	int ret = EXIT_FAILURE;
-	char *config_file = SYSCONFDIR"/openfortivpn/config";
-	char *host, *username = NULL, *password = NULL, *otp = NULL;
+	const char *config_file = SYSCONFDIR"/openfortivpn/config";
+	const char *host;
 	char *port_str;
-	long int port;
 
 	struct vpn_config cfg = {
 		.gateway_host = {'\0'},
-		// gateway_ip
-		.gateway_port = 0,
+		.gateway_port = 443,
 		.username = {'\0'},
-		.password = {'\0'},
+		.password = NULL,
 		.otp = {'\0'},
-		.cookie = {'\0'},
 		.realm = {'\0'},
 		.set_routes = 1,
 		.set_dns = 1,
-		.pppd_use_peerdns = 1,
 		.use_syslog = 0,
 		.half_internet_routes = 0,
+		.persistent = 0,
+#if HAVE_USR_SBIN_PPPD
+		.pppd_use_peerdns = 1,
 		.pppd_log = NULL,
 		.pppd_plugin = NULL,
 		.pppd_ipparam = NULL,
+		.pppd_ifname = NULL,
+		.pppd_call = NULL,
+#endif
+#if HAVE_USR_SBIN_PPP
+		.ppp_system = NULL,
+#endif
 		.ca_file = NULL,
 		.user_cert = NULL,
 		.user_key = NULL,
-		.verify_cert = 1,
 		.insecure_ssl = 0,
 		.cipher_list = NULL,
 		.cert_whitelist = NULL
 	};
+	struct vpn_config cli_cfg = invalid_cfg;
 
 	const struct option long_options[] = {
 		{"help",            no_argument,       0, 'h'},
@@ -175,23 +193,30 @@ int main(int argc, char **argv)
 		{"password",        required_argument, 0, 'p'},
 		{"otp",             required_argument, 0, 'o'},
 		{"set-routes",	    required_argument, 0, 0},
-		{"no-routes",       no_argument, &cfg.set_routes, 0},
+		{"no-routes",       no_argument, &cli_cfg.set_routes, 0},
 		{"half-internet-routes", required_argument, 0, 0},
-		{"set-dns",	    required_argument, 0, 0},
-		{"no-dns",          no_argument, &cfg.set_dns, 0},
-		{"pppd-no-peerdns", no_argument, &cfg.pppd_use_peerdns, 0},
-		{"use-syslog",      no_argument, &cfg.use_syslog, 1},
+		{"set-dns",         required_argument, 0, 0},
+		{"no-dns",          no_argument, &cli_cfg.set_dns, 0},
+		{"use-syslog",      no_argument, &cli_cfg.use_syslog, 1},
+		{"persistent",      required_argument, 0, 0},
 		{"ca-file",         required_argument, 0, 0},
 		{"user-cert",       required_argument, 0, 0},
 		{"user-key",        required_argument, 0, 0},
 		{"trusted-cert",    required_argument, 0, 0},
-		{"insecure-ssl",    no_argument, &cfg.insecure_ssl, 1},
+		{"insecure-ssl",    no_argument, &cli_cfg.insecure_ssl, 1},
 		{"cipher-list",     required_argument, 0, 0},
+#if HAVE_USR_SBIN_PPPD
+		{"pppd-no-peerdns", no_argument, &cli_cfg.pppd_use_peerdns, 0},
 		{"pppd-log",        required_argument, 0, 0},
 		{"pppd-plugin",     required_argument, 0, 0},
 		{"pppd-ipparam",    required_argument, 0, 0},
 		{"pppd-ifname",     required_argument, 0, 0},
+		{"pppd-call",       required_argument, 0, 0},
 		{"plugin",          required_argument, 0, 0}, // deprecated
+#endif
+#if HAVE_USR_SBIN_PPP
+		{"ppp-system",      required_argument, 0, 0},
+#endif
 		{0, 0, 0, 0}
 	};
 
@@ -219,64 +244,77 @@ int main(int argc, char **argv)
 				ret = EXIT_SUCCESS;
 				goto exit;
 			}
+#if HAVE_USR_SBIN_PPPD
 			if (strcmp(long_options[option_index].name,
 			           "pppd-log") == 0) {
-				cfg.pppd_log = strdup(optarg);
+				cli_cfg.pppd_log = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "pppd-plugin") == 0) {
-				cfg.pppd_plugin = strdup(optarg);
+				cli_cfg.pppd_plugin = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "pppd-ifname") == 0) {
-				cfg.pppd_ifname = strdup(optarg);
+				cli_cfg.pppd_ifname = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "pppd-ipparam") == 0) {
-				cfg.pppd_ipparam = strdup(optarg);
-				break;
-			}
-			// --plugin is deprecated, --pppd-plugin should be used
-			if (cfg.pppd_plugin == NULL &&
-			    strcmp(long_options[option_index].name,
-			           "plugin") == 0) {
-				cfg.pppd_plugin = strdup(optarg);
+				cli_cfg.pppd_ipparam = strdup(optarg);
 				break;
 			}
+			if (strcmp(long_options[option_index].name,
+			           "pppd-call") == 0) {
+				cli_cfg.pppd_call = strdup(optarg);
+				break;
+			}
+			// --plugin is deprecated, --pppd-plugin should be used
+			if (cli_cfg.pppd_plugin == NULL &&
+			    strcmp(long_options[option_index].name,
+			           "plugin") == 0) {
+				cli_cfg.pppd_plugin = strdup(optarg);
+				break;
+			}
+#endif
+#if HAVE_USR_SBIN_PPP
+			if (strcmp(long_options[option_index].name,
+			           "ppp-system") == 0) {
+				cfg.ppp_system = strdup(optarg);
+				break;
+			}
+#endif
 			if (strcmp(long_options[option_index].name,
 			           "ca-file") == 0) {
-				cfg.ca_file = strdup(optarg);
+				cli_cfg.ca_file = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "user-cert") == 0) {
-				cfg.user_cert = strdup(optarg);
+				cli_cfg.user_cert = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "user-key") == 0) {
-				cfg.user_key = strdup(optarg);
+				cli_cfg.user_key = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "realm") == 0) {
-				strncpy(cfg.realm, optarg, FIELD_SIZE);
-				cfg.realm[FIELD_SIZE] = '\0';
+				strncpy(cli_cfg.realm, optarg, FIELD_SIZE);
+				cli_cfg.realm[FIELD_SIZE] = '\0';
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "trusted-cert") == 0) {
-				if (add_trusted_cert(&cfg, optarg))
-					log_warn("Could not add certificate "
-					         "digest to whitelist.\n");
+				if (add_trusted_cert(&cli_cfg, optarg))
+					log_warn("Could not add certificate digest to whitelist.\n");
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "cipher-list") == 0) {
-				cfg.cipher_list = strdup(optarg);
+				cli_cfg.cipher_list = strdup(optarg);
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
@@ -287,34 +325,44 @@ int main(int argc, char **argv)
 					         optarg);
 					break;
 				}
-				cfg.set_routes = set_routes;
+				cli_cfg.set_routes = set_routes;
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "half-internet-routes") == 0) {
 				int half_internet_routes = strtob(optarg);
 				if (half_internet_routes < 0) {
-					log_warn("Bad half-internet-routes option: "
-					         "\"%s\"\n", optarg);
+					log_warn("Bad half-internet-routes option: \"%s\"\n",
+					         optarg);
 					break;
 				}
-				cfg.half_internet_routes = half_internet_routes;
+				cli_cfg.half_internet_routes = half_internet_routes;
+				break;
+			}
+			if (strcmp(long_options[option_index].name,
+			           "persistent") == 0) {
+				long int persistent = strtol(optarg, NULL, 0);
+				if (persistent < 0 || persistent > UINT_MAX) {
+					log_warn("Bad persistent option: \"%s\"\n",
+					         optarg);
+					break;
+				}
+				cli_cfg.persistent = persistent;
 				break;
 			}
 			if (strcmp(long_options[option_index].name,
 			           "set-dns") == 0) {
 				int set_dns = strtob(optarg);
 				if (set_dns < 0) {
-					log_warn("Bad set-dns option: \"%s\"\n",
-					         optarg);
+					log_warn("Bad set-dns option: \"%s\"\n", optarg);
 					break;
 				}
-				cfg.set_dns = set_dns;
+				cli_cfg.set_dns = set_dns;
 				break;
 			}
 			goto user_error;
 		case 'h':
-			printf("%s%s%s", usage, help_options, help_config);
+			printf("%s%s%s%s", usage, summary, help_options, help_config);
 			ret = EXIT_SUCCESS;
 			goto exit;
 		case 'v':
@@ -327,13 +375,15 @@ int main(int argc, char **argv)
 			config_file = optarg;
 			break;
 		case 'u':
-			username = optarg;
+			strncpy(cli_cfg.username, optarg, FIELD_SIZE);
+			cli_cfg.username[FIELD_SIZE] = '\0';
 			break;
 		case 'p':
-			password = optarg;
+			cli_cfg.password = strdup(optarg);
 			break;
 		case 'o':
-			otp = optarg;
+			strncpy(cli_cfg.otp, optarg, FIELD_SIZE);
+			cli_cfg.otp[FIELD_SIZE] = '\0';
 			break;
 		default:
 			goto user_error;
@@ -342,55 +392,40 @@ int main(int argc, char **argv)
 
 	if (optind < argc - 1 || optind > argc)
 		goto user_error;
-	set_syslog(cfg.use_syslog);
 
-	if (password != NULL)
-		log_warn("You should not pass the password on the command "
-		         "line. Type it interactively or use a config file "
-		         "instead.\n");
+	if (cli_cfg.password != NULL && cli_cfg.password[0] != '\0')
+		log_warn("You should not pass the password on the command line. Type it interactively or use a config file instead.\n");
+
+	log_debug("openfortivpn " VERSION "\n", config_file);
 
 	// Load config file
 	if (config_file[0] != '\0') {
 		ret = load_config(&cfg, config_file);
-		set_syslog(cfg.use_syslog);
 		if (ret == 0)
 			log_debug("Loaded config file \"%s\".\n", config_file);
 		else
 			log_warn("Could not load config file \"%s\" (%s).\n",
 			         config_file, err_cfg_str(ret));
 	}
+	// Then apply CLI config
+	merge_config(&cfg, &cli_cfg);
+	set_syslog(cfg.use_syslog);
 
 	// Read host and port from the command line
 	if (optind == argc - 1) {
 		host = argv[optind++];
 		port_str = strchr(host, ':');
-		if (port_str == NULL) {
-			log_error("Specify a valid host:port couple.\n");
-			goto user_error;
+		if (port_str != NULL) {
+			port_str[0] = '\0';
+			port_str++;
+			cfg.gateway_port = strtol(port_str, NULL, 0);
+			if (cfg.gateway_port <= 0 || cfg.gateway_port > 65535) {
+				log_error("Specify a valid port.\n");
+				goto user_error;
+			}
 		}
-		port_str[0] = '\0';
 		strncpy(cfg.gateway_host, host, FIELD_SIZE);
 		cfg.gateway_host[FIELD_SIZE] = '\0';
-		port_str++;
-		port = strtol(port_str, NULL, 0);
-		if (port <= 0 || port > 65535) {
-			log_error("Specify a valid port.\n");
-			goto user_error;
-		}
-		cfg.gateway_port = port;
-	}
-	// Read username and password from the command line
-	if (username != NULL) {
-		strncpy(cfg.username, username, FIELD_SIZE);
-		cfg.username[FIELD_SIZE] = '\0';
-	}
-	if (password != NULL) {
-		strncpy(cfg.password, password, FIELD_SIZE);
-		cfg.password[FIELD_SIZE] = '\0';
-	}
-	if (otp != NULL) {
-		strncpy(cfg.otp, otp, FIELD_SIZE);
-		cfg.otp[FIELD_SIZE] = '\0';
 	}
 
 	// Check host and port
@@ -404,9 +439,13 @@ int main(int argc, char **argv)
 		goto user_error;
 	}
 	// If no password given, interactively ask user
-	if (cfg.password[0] == '\0')
-		read_password("VPN account password: ", cfg.password,
-		              FIELD_SIZE);
+	if (cfg.password == NULL || cfg.password[0] == '\0') {
+		free(cfg.password);
+		char *tmp_password = malloc(PWD_BUFSIZ); // allocate large buffer
+		read_password("VPN account password: ", tmp_password, PWD_BUFSIZ);
+		cfg.password = strdup(tmp_password); // copy string of correct size
+		free(tmp_password);
+	}
 	// Check password
 	if (cfg.password[0] == '\0') {
 		log_error("Specify a password.\n");
@@ -421,12 +460,21 @@ int main(int argc, char **argv)
 	if (cfg.otp[0] != '\0')
 		log_debug("One-time password = \"%s\"\n", cfg.otp);
 
-	if (geteuid() != 0)
-		log_warn("This process was not spawned with root "
-		         "privileges, this will probably not work.\n");
+	if (geteuid() != 0) {
+		log_error("This process was not spawned with root privileges, which are required.\n");
+		ret = EXIT_FAILURE;
+		goto exit;
+	}
+
+	do {
+		if (run_tunnel(&cfg) != 0)
+			ret = EXIT_FAILURE;
+		else
+			ret = EXIT_SUCCESS;
+		if ((cfg.persistent > 0) && (get_sig_received() == 0))
+			sleep(cfg.persistent);
+	} while ((get_sig_received() == 0) && (cfg.persistent != 0));
 
-	if (run_tunnel(&cfg) == 0)
-		ret = EXIT_SUCCESS;
 	goto exit;
 
 user_error:
