@@ -34,17 +34,20 @@
 #include <openssl/err.h>
 #include <openssl/ssl.h>
 
-#ifdef __APPLE__
+#ifdef __clang__
 /*
  * Get rid of OSX 10.7 and greater deprecation warnings
  * see for instance https://wiki.openssl.org/index.php/Hostname_validation
  * this pragma selectively suppresses this type of warnings in clang
  */
 #pragma clang diagnostic ignored "-Wdeprecated-declarations"
+#endif
+
+#ifndef ERESTART
 /*
  * ERESTART is one of the recoverable errors which might be returned.
- * However, in OSX this constant is not defined in errno.h so we define
- * a dummy value here.
+ * However, in OSX and BSD this constant is not defined in errno.h
+ * so we define a dummy value here.
  */
 #define ERESTART -1
 #endif
