@@ -39,16 +39,21 @@
 #include <net/if.h>
 #include <arpa/inet.h>
 #include <sys/socket.h>
+#include <sys/ioctl.h>
 #include <openssl/err.h>
 #include <openssl/x509v3.h>
-#ifdef __APPLE__
-#include <util.h>
-#else
+#if HAVE_PTY_H
 #include <pty.h>
+#elif HAVE_UTIL_H
+#include <util.h>
 #endif
+#include <termios.h>
 #include <signal.h>
 #include <sys/wait.h>
 #include <assert.h>
+#if HAVE_SYSTEMD
+#include <systemd/sd-daemon.h>
+#endif
 
 struct ofv_varr {
 	unsigned cap;		// current capacity
@@ -100,6 +105,10 @@ static int on_ppp_if_up(struct tunnel *tunnel)
 
 	log_info("Tunnel is up and running.\n");
 
+#if HAVE_SYSTEMD
+	sd_notify(0, "READY=1");
+#endif
+
 	return 0;
 }
 
@@ -124,7 +133,7 @@ static int pppd_run(struct tunnel *tunnel)
 {
 	pid_t pid;
 	int amaster;
-#ifndef __APPLE__
+#ifdef HAVE_STRUCT_TERMIOS
 	struct termios termp = {
 		.c_cflag = B9600,
 		.c_cc[VTIME] = 0,
@@ -132,32 +141,48 @@ static int pppd_run(struct tunnel *tunnel)
 	};
 #endif
 
-	static const char pppd_path[] = "/usr/sbin/pppd";
-
-	if (access(pppd_path, F_OK) != 0) {
-		log_error("%s: %s.\n", pppd_path, strerror(errno));
+	static const char ppp_path[] = PPP_PATH;
+	if (access(ppp_path, F_OK) != 0) {
+		log_error("%s: %s.\n", ppp_path, strerror(errno));
 		return 1;
 	}
+	log_debug("ppp_path: %s\n", ppp_path);
 
-#ifdef __APPLE__
-	pid = forkpty(&amaster, NULL, NULL, NULL);
-#else
+#ifdef HAVE_STRUCT_TERMIOS
 	pid = forkpty(&amaster, NULL, &termp, NULL);
+#else
+	pid = forkpty(&amaster, NULL, NULL, NULL);
 #endif
 
 	if (pid == -1) {
 		log_error("forkpty: %s\n", strerror(errno));
 		return 1;
 	} else if (pid == 0) { // child process
+
 		struct ofv_varr pppd_args = { 0, 0, NULL };
 
+#if HAVE_USR_SBIN_PPP
+		/*
+		* assume there is a default configuration to start.
+		* Support for taking options from the command line
+		* e.g. the name of the configuration or options
+		* to send interactively to ppp will be added later
+		*/
+		const char *v[] = {
+			ppp_path,
+			"-direct"
+		};
+		for (unsigned i = 0; i < sizeof v/sizeof v[0]; i++)
+			ofv_append_varr(&pppd_args, v[i]);
+#endif
+#if HAVE_USR_SBIN_PPPD
 		if (tunnel->config->pppd_call) {
-			ofv_append_varr(&pppd_args, pppd_path);
+			ofv_append_varr(&pppd_args, ppp_path);
 			ofv_append_varr(&pppd_args, "call");
 			ofv_append_varr(&pppd_args, tunnel->config->pppd_call);
 		} else {
 			const char *v[] = {
-				pppd_path,
+				ppp_path,
 				"38400", // speed
 				":192.0.2.1", // <local_IP_address>:<remote_IP_address>
 				"noipdefault",
@@ -174,7 +199,6 @@ static int pppd_run(struct tunnel *tunnel)
 			for (unsigned i = 0; i < ARRAY_SIZE(v); i++)
 				ofv_append_varr(&pppd_args, v[i]);
 		}
-
 		if (tunnel->config->pppd_use_peerdns)
 			ofv_append_varr(&pppd_args, "usepeerdns");
 		if (tunnel->config->pppd_log) {
@@ -194,10 +218,17 @@ static int pppd_run(struct tunnel *tunnel)
 			ofv_append_varr(&pppd_args, "ifname");
 			ofv_append_varr(&pppd_args, tunnel->config->pppd_ifname);
 		}
+#endif
+#if HAVE_USR_SBIN_PPP
+		if (tunnel->config->ppp_system) {
+			ofv_append_varr(&pppd_args, tunnel->config->ppp_system);
+		}
+#endif
 
 		close(tunnel->ssl_socket);
 		execv(pppd_args.data[0], (char *const *)pppd_args.data);
 		free(pppd_args.data);
+
 		/*
 		 * The following call to fprintf() doesn't work, probably
 		 * because of the prior call to forkpty().
@@ -249,7 +280,8 @@ static int pppd_terminate(struct tunnel *tunnel)
 {
 	close(tunnel->pppd_pty);
 
-	log_debug("Waiting for pppd to exit...\n");
+	log_debug("Waiting for %s to exit...\n", PPP_DAEMON);
+
 	int status;
 	if (waitpid(tunnel->pppd_pid, &status, 0) == -1) {
 		log_error("waitpid: %s\n", strerror(errno));
@@ -257,28 +289,52 @@ static int pppd_terminate(struct tunnel *tunnel)
 	}
 	if (WIFEXITED(status)) {
 		int exit_status = WEXITSTATUS(status);
-		log_debug("waitpid: pppd exit status code %d\n", exit_status);
-		if (exit_status >= ARRAY_SIZE(pppd_message) || exit_status < 0)
-			log_error("pppd: Returned an unknown exit status: %d\n",
-			          exit_status);
-		else
+		log_debug("waitpid: %s exit status code %d\n",
+		          PPP_DAEMON, exit_status);
+#if HAVE_USR_SBIN_PPPD
+		if (exit_status >= ARRAY_SIZE(pppd_message) || exit_status < 0) {
+			log_error("%s: Returned an unknown exit status: %d\n",
+			          PPP_DAEMON, exit_status);
+		} else {
 			switch (exit_status) {
 			case 0: // success
-				log_debug("pppd: %s\n", pppd_message[exit_status]);
+				log_debug("%s: %s\n",
+				          PPP_DAEMON, pppd_message[exit_status]);
 				break;
 			case 16: // emitted when exiting normally
-				log_info("pppd: %s\n", pppd_message[exit_status]);
+				log_info("%s: %s\n",
+				         PPP_DAEMON, pppd_message[exit_status]);
 				break;
 			default:
-				log_error("pppd: %s\n", pppd_message[exit_status]);
+				log_error("%s: %s\n",
+				          PPP_DAEMON, pppd_message[exit_status]);
 				break;
 			}
+		}
+#else
+		// ppp exit codes in the FreeBSD case
+		switch (exit_status) {
+		case 0: // success and EX_NORMAL as defined in ppp source directly
+			log_debug("%s: %s\n", PPP_DAEMON, pppd_message[exit_status]);
+			break;
+		case 1:
+		case 127:
+		case 255: // abnormal exit with hard-coded error codes in ppp
+			log_error("%s: exited with return value of %d\n",
+			          PPP_DAEMON, exit_status);
+			break;
+		default:
+			log_error("%s: %s (%d)\n", PPP_DAEMON, strerror(exit_status),
+			          exit_status);
+			break;
+		}
+#endif
 	} else if (WIFSIGNALED(status)) {
 		int signal_number = WTERMSIG(status);
-		log_debug("waitpid: pppd terminated by signal %d\n",
-		          signal_number);
-		log_error("pppd: terminated by signal: %s\n",
-		          strsignal(signal_number));
+		log_debug("waitpid: %s terminated by signal %d\n",
+		          PPP_DAEMON, signal_number);
+		log_error("%s: terminated by signal: %s\n",
+		          PPP_DAEMON, strsignal(signal_number));
 	}
 
 	return 0;
@@ -296,10 +352,17 @@ int ppp_interface_is_up(struct tunnel *tunnel)
 	}
 
 	for (ifa = ifap; ifa != NULL; ifa = ifa->ifa_next) {
-		if (((tunnel->config->pppd_ifname
-		      && strstr(ifa->ifa_name, tunnel->config->pppd_ifname) != NULL)
-		     || strstr(ifa->ifa_name, "ppp") != NULL)
-		    && ifa->ifa_flags & IFF_UP) {
+		if ((
+#if HAVE_USR_SBIN_PPPD
+		            (tunnel->config->pppd_ifname
+		             && strstr(ifa->ifa_name, tunnel->config->pppd_ifname)
+		             != NULL)
+		            || strstr(ifa->ifa_name, "ppp") != NULL)
+#endif
+#if HAVE_USR_SBIN_PPP
+		    strstr(ifa->ifa_name, "tun") != NULL)
+#endif
+			&& ifa->ifa_flags& IFF_UP) {
 			if (&(ifa->ifa_addr->sa_family) != NULL
 			    && ifa->ifa_addr->sa_family == AF_INET) {
 				struct in_addr if_ip_addr =
@@ -725,9 +788,12 @@ int ssl_connect(struct tunnel *tunnel)
 	if (!tunnel->config->insecure_ssl && !tunnel->config->cipher_list) {
 		const char *cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
 
-		if (tunnel->config->cipher_list)
-			cipher_list = tunnel->config->cipher_list;
-		if (!SSL_set_cipher_list(tunnel->ssl_handle, cipher_list)) {
+		tunnel->config->cipher_list = strdup(cipher_list);
+	}
+
+	if (tunnel->config->cipher_list) {
+		if (!SSL_set_cipher_list(tunnel->ssl_handle,
+		                         tunnel->config->cipher_list)) {
 			log_error("SSL_set_cipher_list failed: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
 			return 1;
@@ -750,9 +816,8 @@ int ssl_connect(struct tunnel *tunnel)
 	}
 	SSL_set_mode(tunnel->ssl_handle, SSL_MODE_AUTO_RETRY);
 
-	if (tunnel->config->verify_cert)
-		if (ssl_verify_cert(tunnel))
-			return 1;
+	if (ssl_verify_cert(tunnel))
+		return 1;
 
 	// Disable SIGPIPE (occurs when trying to write to an already-closed
 	// socket).
@@ -776,11 +841,13 @@ int run_tunnel(struct vpn_config *config)
 	};
 
 	// Step 0: get gateway host IP
+	log_debug("Resolving gateway host ip\n");
 	ret = get_gateway_host_ip(&tunnel);
 	if (ret)
 		goto err_tunnel;
 
 	// Step 1: open a SSL connection to the gateway
+	log_debug("Establishing ssl connection\n");
 	ret = ssl_connect(&tunnel);
 	if (ret)
 		goto err_tunnel;
@@ -790,13 +857,13 @@ int run_tunnel(struct vpn_config *config)
 	// cookie
 	ret = auth_log_in(&tunnel);
 	if (ret != 1) {
-		log_error("Could not authenticate to gateway (%s).\n",
-		          err_http_str(ret));
+		log_error("Could not authenticate to gateway. Please check the password, client certificate, etc.\n");
+		log_debug("%s %d\n", err_http_str(ret), ret);
 		ret = 1;
 		goto err_tunnel;
 	}
 	log_info("Authenticated.\n");
-	log_debug("Cookie: %s\n", config->cookie);
+	log_debug("Cookie: %s\n", tunnel.cookie);
 
 	ret = auth_request_vpn_allocation(&tunnel);
 	if (ret != 1) {
@@ -812,6 +879,7 @@ int run_tunnel(struct vpn_config *config)
 		goto err_tunnel;
 
 	// Step 3: get configuration
+	log_debug("Retrieving configuration\n");
 	ret = auth_get_config(&tunnel);
 	if (ret != 1) {
 		log_error("Could not get VPN configuration (%s).\n",
@@ -821,16 +889,18 @@ int run_tunnel(struct vpn_config *config)
 	}
 
 	// Step 4: run a pppd process
+	log_debug("Establishing the tunnel\n");
 	ret = pppd_run(&tunnel);
 	if (ret)
 		goto err_tunnel;
 
 	// Step 5: ask gateway to start tunneling
+	log_debug("Switch to tunneling mode\n");
 	ret = http_send(&tunnel,
 	                "GET /remote/sslvpn-tunnel HTTP/1.1\r\n"
 	                "Host: sslvpn\r\n"
 	                "Cookie: %s\r\n\r\n",
-	                tunnel.config->cookie);
+	                tunnel.cookie);
 	if (ret != 1) {
 		log_error("Could not start tunnel (%s).\n", err_http_str(ret));
 		ret = 1;
@@ -841,8 +911,10 @@ int run_tunnel(struct vpn_config *config)
 	ret = 0;
 
 	// Step 6: perform io between pppd and the gateway, while tunnel is up
+	log_debug("Starting IO through the tunnel\n");
 	io_loop(&tunnel);
 
+	log_debug("disconnecting\n");
 	if (tunnel.state == STATE_UP)
 		if (tunnel.on_ppp_if_down != NULL)
 			tunnel.on_ppp_if_down(&tunnel);
@@ -851,7 +923,7 @@ int run_tunnel(struct vpn_config *config)
 
 err_start_tunnel:
 	pppd_terminate(&tunnel);
-	log_info("Terminated pppd.\n");
+	log_info("Terminated %s.\n", PPP_DAEMON);
 err_tunnel:
 	log_info("Closed connection to gateway.\n");
 	tunnel.state = STATE_DOWN;
