@@ -32,11 +32,14 @@
 #include "config.h"
 #include "io.h"
 #include "ipv4.h"
+#if HAVE_LIBUTIL_H
+#include "libutil.h"
+#endif
 
 #include <openssl/ssl.h>
 #include <sys/types.h>
 
-#ifdef __APPLE__
+#ifdef __clang__
 /*
  * Get rid of OSX 10.7 and greater deprecation warnings
  * see for instance https://wiki.openssl.org/index.php/Hostname_validation
@@ -56,6 +59,7 @@ struct tunnel {
 	struct vpn_config *config;
 
 	enum tunnel_state state;
+	char cookie[COOKIE_SIZE + 1];
 
 	struct ppp_packet_pool ssl_to_pty_pool;
 	struct ppp_packet_pool pty_to_ssl_pool;
